#include "src/ui/core-installer.hpp"

#include <QMessageBox>
#include <QDateTime>
#include <QCoreApplication>
#include <cassert>
#include <thread>
#include "src/polyfill/base-directory.hpp"
#include "src/ui/download-dialog.hpp"
#include "src/ui/core-finder-dialog.hpp"
#include "src/core/file-controller.hpp"
#include "src/core/retroarch.hpp"
#include "src/core/time.hpp"
#include "src/core/zip.hpp"

static string installCoreSync( const CoreBuild &coreBuild ) {
	const fs::path tempPath = BaseDir::temp() / "core.zip";
	fs::create_directories( BaseDir::temp() );

	const DownloadResult status = DownloadDialog::download( QT_TRANSLATE_NOOP( "DownloadDialog", "Downloading core..." ), coreBuild.downloadLink, tempPath );
	if( !status.success ) {
		QMessageBox::critical(
			nullptr,
			QCoreApplication::translate( "CoreInstaller", "Download Failed" ),
			QCoreApplication::translate( "CoreInstaller", "Failed to download emulator core" ).append( '\n' ).append( status.errorMessage.c_str() )
		);
		return "";
	}

	if( !Zip::unzip( tempPath, RetroArch::getCorePath() ) ) {
		QMessageBox::critical(
			nullptr,
			QCoreApplication::translate( "CoreInstaller", "Installation Failed" ),
			QCoreApplication::translate( "CoreInstaller", "Failed to install emulator core" )
		);
		return "";
	}

	fs::error_code err;
	fs::remove( tempPath, err );

	QMessageBox::information(
		nullptr,
		QCoreApplication::translate( "CoreInstaller", "Installation Successful" ),
		QCoreApplication::translate( "CoreInstaller", "Core installed successfully." )
	);
	return coreBuild.commit.hash;
}

bool CoreInstaller::requireCore( EmulatorCore core ) {
	if( core == EmulatorCore::UseDefault ) {
		core = FileController::loadAppSettings().defaultEmulator;
	}

	if( RetroArch::isEmulatorCoreInstalled( core ) ) {
		return true;
	}

	assert( core == EmulatorCore::ParallelN64 || core == EmulatorCore::Mupen64plusNext );

#ifdef __APPLE__
	if( core == EmulatorCore::Mupen64plusNext ) {
		QMessageBox::critical(
			nullptr,
			QCoreApplication::translate( "CoreInstaller", "Core Unavailable" ),
			QCoreApplication::translate( "CoreInstaller", "Sorry! The Mupen64plus-Next emulator core is not available on MacOS. Try using ParallelN64 instead." )
		);
		return false;
	}
#endif

	if( QMessageBox::question(
		nullptr,
		QCoreApplication::translate( "CoreInstaller", "Install Emulator Core?" ),
		QCoreApplication::translate( "CoreInstaller", "This emulator core is not installed. Would you like to install it now?" )
	) != QMessageBox::Yes ) {
		return false;
	}

	CoreFinderDialog coreFinder( core );
	coreFinder.exec();

	const std::optional<CoreBuild> &coreBuild = coreFinder.coreBuild();
	if( coreBuild.has_value() ) {
		return !installCoreSync( coreBuild.value() ).empty();
	}

	return false;
}

static void checkForUpdates(
	const InstalledVersionsInfo &currentVersionInfo,
	EmulatorCore core,
	std::function<void(string)> then
) {
	const string currentCommit = (core == EmulatorCore::ParallelN64) ? currentVersionInfo.parallelVersion.commit : currentVersionInfo.mupenVersion.commit;

#ifdef __APPLE__
	if( core == EmulatorCore::Mupen64plusNext ) {
		then( currentCommit );
		return;
	}
#endif

	if( !RetroArch::isEmulatorCoreInstalled( core ) ) {
		then( currentCommit );
		return;
	}

	const char *coreName;
	string branch;
	if( core == EmulatorCore::ParallelN64 ) {
		if( currentVersionInfo.parallelVersion.lock ) {
			then( currentCommit );
			return;
		}
		coreName = "ParallelN64";
		branch = "master";
	} else {
		if( currentVersionInfo.mupenVersion.lock ) {
			then( currentCommit );
			return;
		}
		coreName = "Mupen64Plus-Next";
		branch = FileController::loadAppSettings().mupenDevBranch ? "develop" : "master";
	}

	CoreBuilds::getLastKnownGood(
		core,
		branch,
		currentCommit,
		[=](const CoreBuild &build) {
			if( QMessageBox::question(
				nullptr,
				QCoreApplication::translate( "CoreInstaller", "Core Update Available" ),
				QCoreApplication::translate( "CoreInstaller", "An update is available for %1. Would you like to install it now?" ).arg( coreName )
			) == QMessageBox::Yes ) {
				const string commit = installCoreSync( build );
				if( !commit.empty() ) {
					InstalledVersionsInfo versions = FileController::loadInstalledVersions();
					if( core == EmulatorCore::ParallelN64 ) {
						versions.parallelVersion.commit = commit;
					} else {
						versions.mupenVersion.commit = commit;
					}
					FileController::saveInstalledVersions( versions );
					then( commit );
					return;
				}
			}
			then( currentCommit );
		},
		[=](){ then( currentCommit ); }
	);
}

void CoreInstaller::checkForUpdatesAsync() {
	const InstalledVersionsInfo versions = FileController::loadInstalledVersions();
	checkForUpdates(
		versions,
		EmulatorCore::ParallelN64,
		[=](string parallelVersion) {
#ifdef __APPLE__
			InstalledVersionsInfo updatedVersions = versions;
			updatedVersions.parallelVersion.commit = parallelVersion;
			updatedVersions.lastUpdateCheck = Time::now();
			FileController::saveInstalledVersions( updatedVersions );
#else
			QMetaObject::invokeMethod( QCoreApplication::instance(), [=]() {
				checkForUpdates(
					versions,
					EmulatorCore::Mupen64plusNext,
					[=](string mupenVersion){
						InstalledVersionsInfo updatedVersions = versions;
						updatedVersions.parallelVersion.commit = parallelVersion;
						updatedVersions.mupenVersion.commit = mupenVersion;
						updatedVersions.lastUpdateCheck = Time::now();
						FileController::saveInstalledVersions( updatedVersions );
					}
				);
			}, Qt::QueuedConnection );
#endif
		}
	);
}

void CoreInstaller::checkForUpdatesSync( InstalledVersionsInfo &versions ) {
	bool done = false;
	checkForUpdates(
		versions,
		EmulatorCore::ParallelN64,
		[&](string commit) {
			versions.parallelVersion.commit = commit;
#ifdef __APPLE__
			done = true;
#else
			QMetaObject::invokeMethod( QCoreApplication::instance(), [&]() {
				checkForUpdates(
					versions,
					EmulatorCore::Mupen64plusNext,
					[&](string commit){
						versions.mupenVersion.commit = commit;
						versions.lastUpdateCheck = Time::now();
						done = true;
					}
				);
			}, Qt::QueuedConnection );
#endif
		}
	);

	while( !done ) {
		QCoreApplication::processEvents( QEventLoop::ExcludeUserInputEvents );
		std::this_thread::yield();
	}
}
