#include "src/windows-mac-updater/retroarch-updater.hpp"

#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QProgressDialog>
#include <QMessageBox>
#include <QCoreApplication>
#include <functional>
#include <memory>
#include <thread>
#include "src/core/qthread.hpp"
#include "src/core/version.hpp"
#include "src/core/file-controller.hpp"
#include "src/core/retroarch.hpp"
#include "src/polyfill/base-directory.hpp"
#include "src/polyfill/process.hpp"
#include "src/ui/download-dialog.hpp"
#ifdef __APPLE__
#include "src/core/uuid.hpp"
#endif

static void headAsync(
	QNetworkAccessManager *web,
	const string &url,
	const std::function<void(bool)> &callback
) {
	QNetworkReply *response = web->head( QNetworkRequest( QUrl( url.c_str() ) ) );

	QObject::connect( response, &QNetworkReply::finished, web, [=](){
		if( response->error() != QNetworkReply::NoError ) {
			callback( false );
			return;
		}

		bool hasStatusCode = false;
		const int statusCode = response->attribute( QNetworkRequest::HttpStatusCodeAttribute ).toInt( &hasStatusCode );
		response->deleteLater();
		callback( hasStatusCode && statusCode == 200 );
	}, Qt::QueuedConnection );
}

enum VersionMarch {
	Patch = 0,
	Minor = 1,
	Major = 2
};

static inline string buildDownloadUrl( const Version &version ) {
#ifdef __APPLE__
	return "https://buildbot.libretro.com/stable/"s + version.toString() + "/apple/osx/x86_64/RetroArch.dmg";
#else
	#ifdef _WIN64
		return "https://buildbot.libretro.com/stable/"s + version.toString() + "/windows/x86_64/RetroArch.7z";
	#else
		return "https://buildbot.libretro.com/stable/"s + version.toString() + "/windows/x86/RetroArch.7z";
	#endif
#endif
}

static inline Version marchVersion(
	Version version,
	VersionMarch versionMarch
) {
	switch( versionMarch ) {
		case VersionMarch::Patch: version.patch++; break;
		case VersionMarch::Minor: version.minor++; version.patch = 0; break;
		case VersionMarch::Major: version.major++; version.minor = 0; version.patch = 0; break;
		default: break;
	}
	return version;
}

static const Version EARLIEST_VERSION = Version{ 1, 9, 0 };

static void getWindowsDownloadHelper(
	QNetworkAccessManager *web,
	Version lastFoundVersion,
	Version earliestVersion,
	VersionMarch versionMarch,
	std::shared_ptr<bool> cancelled,
	std::function<void(Version)> callback
) {
	if( *cancelled ) {
		callback( EARLIEST_VERSION );
		return;
	}

	const string url = buildDownloadUrl( earliestVersion ).c_str();
	headAsync(
		web,
		url,
		[=](bool found) {
			if( found ) {
				getWindowsDownloadHelper( web, earliestVersion, marchVersion( earliestVersion, VersionMarch::Patch ), VersionMarch::Patch, cancelled, callback );
			} else if( versionMarch == VersionMarch::Major ) {
				callback( lastFoundVersion );
			} else {
				const VersionMarch newVersionMarch = (VersionMarch)((int)versionMarch + 1);
				getWindowsDownloadHelper( web, lastFoundVersion, marchVersion( earliestVersion, newVersionMarch ), newVersionMarch, cancelled, callback );
			}
		}
	);
}

static inline Version getLatestWindowsVersion(
	const Version &earliestVersion
) {
	QProgressDialog dialog;
	dialog.setRange( 0, 0 );
	dialog.setLabelText( QCoreApplication::translate( "WindowsRetroArchUpdater", "Checking for RetroArch Updates" ) );

	std::shared_ptr<bool> cancelled( new bool( false ) );
	Version latestVersion = earliestVersion;

	QNetworkAccessManager *web = new QNetworkAccessManager();
	getWindowsDownloadHelper( web, earliestVersion, earliestVersion, VersionMarch::Patch, cancelled, [cancelled,&latestVersion,&dialog](Version foundVersion){
		if( *cancelled ) return;
		latestVersion = foundVersion;
		dialog.close();
	});

	dialog.exec();
	*cancelled = true;
	web->deleteLater();
	return latestVersion;
}

#ifdef _WIN64
#define RA_ROOT "RetroArch-Win64"
#else
#define RA_ROOT "RetroArch-Win32"
#endif

static void installWorker(
	QProgressDialog *dialog,
	std::shared_ptr<bool> cancelled,
	bool *result
) {
	const fs::path installDir = RetroArch::getExePath().parent_path();

	fs::error_code err;
	fs::remove_all( installDir, err );
	fs::create_directories( installDir, err );

#ifdef _WIN32
	AsyncProcess extraction(
		(BaseDir::program() / "7za.exe").u8string(),
		{
			"x", (BaseDir::temp() / "retroarch.7z").u8string(),
			"-t7z",
			"-aoa",
			"-y",
			"-o"s + installDir.u8string()
		}
	);
	const uint exitCode = extraction.join();

	if( exitCode == 0 ) {
		if( fs::isDirectorySafe( installDir / RA_ROOT ) ) {
			const fs::path tempDir = installDir.parent_path() / "retroarch-temp";
			fs::remove_all( tempDir, err );
			fs::rename( installDir / RA_ROOT, tempDir, err );
			fs::remove_all( installDir, err );
			fs::rename( tempDir, installDir, err );
		}
	}

	QtThread::safeAsync( [=](){
		if( *cancelled ) return;
		*result = (exitCode == 0 && fs::existsSafe( RetroArch::getExePath() ));
		dialog->close();
	});
#else
	const fs::path tempMountPath = BaseDir::temp() / Uuid::Random().toString();
	AsyncProcess mountProcess(
		"hdiutil",
		{
			"attach",
			(BaseDir::temp() / "RetroArch.dmg").u8string(),
			"-mountpoint",
			tempMountPath.u8string()
		}
	);

	const uint exitCode = mountProcess.join();
	if( exitCode != 0 ) {
		QtThread::safeAsync( [=](){
			if( *cancelled ) return;
			*result = false;
			dialog->close();
		});
		return;
	}

	fs::remove_all( installDir / "RetroArch.app", err );
	*result = fs::tryCopyRecursive(
		tempMountPath / "RetroArch.app",
		installDir / "RetroArch.app"
	);

	AsyncProcess(
		"hdiutil",
		{
			"detach",
			tempMountPath.u8string()
		}
	).join();

	RetroArch::installAssets();

	QtThread::safeAsync( [=](){
		if( *cancelled ) return;
		dialog->close();
	});
#endif
}

static inline bool downloadAndInstall( const Version &version ) {
#ifdef __APPLE__
	const fs::path tempBundlePath = BaseDir::temp() / "RetroArch.dmg";
#else
	const fs::path tempBundlePath = BaseDir::temp() / "retroarch.7z";
#endif

	fs::error_code err;
	fs::remove( tempBundlePath, err );

	const DownloadResult status = DownloadDialog::download(
		QT_TRANSLATE_NOOP( "DownloadDialog", "Downloading RetroArch" ),
		buildDownloadUrl( version ),
		tempBundlePath
	);

	if( !status.success ) {
		QMessageBox::critical(
			nullptr,
			QCoreApplication::translate( "WindowsRetroArchUpdater", "Download Failed" ),
			QCoreApplication::translate( "WindowsRetroArchUpdater", "Failed to download RetroArch." ).append( '\n' ).append( status.errorMessage.c_str() )
		);
		return false;
	}

	QProgressDialog dialog;
	dialog.setRange( 0, 0 );
	dialog.setLabelText( QCoreApplication::translate( "WindowsRetroArchUpdater", "Installing RetroArch" ) );

	bool success = false;
	std::shared_ptr<bool> cancelled( new bool( false ) );

	std::thread( installWorker, &dialog, cancelled, &success ).detach();
	dialog.exec();
	*cancelled = true;

	if( !success ) {
		QMessageBox::critical(
			nullptr,
			QCoreApplication::translate( "WindowsRetroArchUpdater", "Installation Error" ),
			QCoreApplication::translate( "WindowsRetroArchUpdater", "An error occurred attempting to uncompress the portable RetroArch bundle" )
		);
	}

	return success;
}

bool RetroArchUpdater::install() {
	const Version latestVersion = getLatestWindowsVersion( EARLIEST_VERSION );
	if( latestVersion < EARLIEST_VERSION ) {
		return false;
	}

	if( downloadAndInstall( latestVersion ) ) {
		InstalledVersionsInfo versions = FileController::loadInstalledVersions();
		versions.retroArchVersion.version = latestVersion;
		FileController::saveInstalledVersions( versions );
		return true;
	}

	return false;
}

void RetroArchUpdater::update() {
	InstalledVersionsInfo versions = FileController::loadInstalledVersions();
	Version &version = versions.retroArchVersion.version;
	if( version < EARLIEST_VERSION ) {
		version = EARLIEST_VERSION;
	}

	Version latestVersion = getLatestWindowsVersion( version );
	if( latestVersion <= version ) {
		return;
	}

	if( QMessageBox::question(
		nullptr,
		QCoreApplication::translate( "WindowsRetroArchUpdater", "Install Update?" ),
		QCoreApplication::translate( "WindowsRetroArchUpdater", "An update for RetroArch is available. Would you like to install it now?" )
	) != QMessageBox::Yes ) {
		return;
	}

	version = latestVersion;
	if( downloadAndInstall( latestVersion ) ) {
		FileController::saveInstalledVersions( versions );
	}
}
